import {sleep, __ as cls, _string, _array, _primitive, _boolean, _number, Try, _object, no_emptyObject as neo} from "../utils/lib.js";

const validator = function(details) {
    let errMessage = null;
    let allIds = [];
    details.forEach((detail, index) => {
        let errorTrace = `Error Trace ${JSON.stringify(detail)}`;
        if (typeof detail !== 'object') {
            errMessage = `Invalid data found. ${errorTrace}`;
        } else if (typeof detail.id !== 'string' || !detail.id.match(/^[_a-z]/)) {
            errMessage = `'id' is required for every field and must start with a letter or underscore (_). ${errorTrace}`;
        } else {
            if(allIds.includes(detail.id))
                errMessage = `Unique 'id' is required for every field, Same 'id' detected, please double check the fields. ${errorTrace}`;
            allIds.push(detail.id)
            for (let i in detail) {
                if (i === 'validate' && detail[i]) {
                    let vd = detail[i];
                    if (vd && typeof vd !== 'object') {
                        errMessage = `Validate must be an object. ${errorTrace}`
                    } else {
                        let vtor = vd.validator;
                        if (!vtor)
                            errMessage = `If validate is provided, validator must also be provided. ${errorTrace}`;
                        else {
                            if (typeof vtor !== 'function')
                                errMessage = `Validator must be a function. ${errorTrace}`;
                        }
                    }
                } else if (i === 'required' && typeof detail[i] !== 'boolean')
                    errMessage = `required property must be of type boolean. ${errorTrace}`;
                else {
                    let type = detail.type;
                    let opts = detail.options;
                    let dv = detail.default_value;
                    if (i === 'default_value' && dv && type === 'number' && typeof dv !== 'number') {
                        errMessage = `If type is number, default_value must be of same type. ${errorTrace}`;
                    } else if (i === 'options') {
                        if (!Array.isArray(opts) ||
                            !opts.every(opt => Object.values(opt).every(e => !['function', 'object'].includes(typeof e))) ||
                            !opts.every(opt => ['id', 'value'].every(e => Object.keys(opt).includes(e))))
                            errMessage = `'options' should be an array of Objects having 'id' and 'value' properties, all of primitive types. ${errorTrace}`;
                    } else if (['min', 'max'].includes(i) && typeof detail[i] !== 'number')
                        errMessage = `'${i}' must be a number. ${errorTrace}`;
                    else {
                        if (['type', 'placeholder', 'description', 'label', 'id'].includes(i) &&
                            typeof detail[i] !== 'string')
                            errMessage = `Invalid value encountered. ${errorTrace}`;
                    }
                }

            };
        }
    })
    return { error: errMessage && `FormValidationError: ${errMessage}` || errMessage }
}

let __;
let _;

export default function Form(jql, info) {
       let ok, launched, _static, validating, modal;
        _ = jql;
        __ = (className, tagName)=> cls.call(_, className, tagName);
        info = _object(info);
        let form = _('<form class = "needs-validation jquery-lite-form"/>');
        let attr = info.attributes;
        if(attr)
          form.attr(attr);
        let loading = _('< class = "d-flex align-items-center text-secondary"/>').addChild([
            _('<strong hidden>Validating...</strong>'),
            _('< hidden class = "spinner-border ms-auto"/>')
        ]);
        let fields = [];
        let st = info.submitText;
        let container = _('</>').addChild([loading, _('< class="mb-1"/>').addChild(info.description || ''), form]);
        let submitButton = _(`<button class = "btn btn-primary submit-btn">${typeof st === 'string' ? st : 'Done'}</button>`).click(submit);
        let cancelButton = _('<button class = "btn btn-light">Cancel</button>').click(() => modal && modal.hide());

   function checkInput(info, value){
    let error = null;
    return new Promise(async(res) => {
        if ((info.required && String(value).trim() === '') || (info.type === 'checkbox' && value == false))
            error = 'This field is required';
        if (info.validate) {
            error = _object(await Try(() => info.validate.validator(value))).error;
        }
          res(error);
    })
   }

   function getSpecificDetails(field){
    let type = field.attr('type');
    let value = type === 'checkbox' ? field.plain().checked : (type == 'file' ? [].slice.call(field.plain().files) : field.value())
    let info = fields.find(i => i.id === field.id());
    return { value, info };
   }

   function handleImmidateErrors(field){
        field.on('keyup cut paste change', async() => {
            validating = true;
            await sleep(100);
            let sd = getSpecificDetails(field);
            loader().show();
            let error = await checkInput(sd.info, sd.value);
            let _err = field.parent().find('.text-danger');
            if (error) {
                field.addClass('invalid').removeClass('valid');
                _err.refill(error);
            } else {
                field.removeClass('invalid').addClass('valid');
                _err.truncate();
            }
            loader().hide();
            validating = false;
        });
    };
    function loader(){
        let action = hide => loading.children().get().forEach(c => c.hide(hide));
        return {
            show: () => action(false),
            hide: () => action(),
            message: message =>{
                 loading.firstChild().truncate().text(message)
                 return loader();
            }
        }
    }

     function addFields(_fields, after){
        _fields = _array(_fields, [_fields]);
        let updatedFields = [...fields, ..._fields];
        let { error } = validator(updatedFields);
        if (error) return console.error(error);
        if(launched === false) return console.error('Form instance does\'t exist');
        fields = updatedFields;

        _fields.forEach(detail => {
            let type = detail.type || 'text';
            if (type === 'submit')
                type = 'text';
            if (type === 'radio')
                type = 'checkbox';
            let id = detail.id;
            let placeholder = _primitive(detail.placeholder);
            let default_value = _primitive(detail.default_value);
            let label = _primitive(detail.label);
            let description = _primitive(detail.description);
            let required = _boolean(detail.required, false);
            let options_ = _array(detail.options);
            let min = _number(detail.min);
            let max = _number(detail.max, 100);
            let vd = detail.validate;
            let _when = _string(vd && vd.when || '').toLowerCase();
            let when = (['live', 'submit'].includes(_when)) ? _when : 'submit';
            let attr = {type, id, name: id, title: label};
            let _label = label ? _('<label/>').addClass('form-label h6').attr({ for: id }).addChild(label) : _('<span />');
            let _input = _('<input class = "form-control shadow-none"/>')
                .required(required).attr(neo({
                placeholder,
                value: default_value,
                 ...attr}));
    
            let _description = _('< class="text-secondary"/>').addChild(description);
            let _error = _('< class="text-danger"/>');
    
            let _container = _(`< class="mt-2" id="field_${id}" field-id = "${id}"/>`);
            if (type === 'checkbox') {
                _input = _('<input class="form-check-input shadow-none"/>').required(required).attr(neo(attr))
                _label.className('form-check-label');
                _container.addClass('form-check form-switch').addChild([_input, _label, _error, _description]);
            } else if (type === 'list') {
                let listId = 'dataList_' + id;
                let dList = _('<datalist/>').id(listId).addChild(options_.map(op => _('<option/>').addChild(op.value)));
                _input.removeAttribute('type').attr({ list: listId });
                _container.addChild([_label, _input, dList, _description, _error]);
            } else if (type === 'select') {
                _input = _(`<select class="form-select shadow-none"/>`)
                         .addChild(options_.map(op => _('<option/>').selected(op.selected).addChild(op.value)))
                         .attr(neo(attr))
                _container.addChild([_label, _input, _description, _error]);
            } else if (type === 'textarea') {
                _input = _('<textarea class = "form-control shadow-none"/>').attr(neo({
                    placeholder,
                    ...attr
                }));
                _input.text(default_value);
                _container.addChild([_label, _input, _description, _error]);
            } else {
                if (['number', 'range'].includes(type))
                    _input.attr({ min, max });
                _container.addChild([_label, _input, _description, _error]);
            }
            if(after){
              let _after = form.find(`#field_${after}`);
    
              if(_after.plain() && _after.siblings().get(1).plain(0))
                   form.insertBefore(_container, _after.siblings().get(1))
              else
                  form.addChild(_container);
        
            }else{
                form.addChild(_container);
            }
    
            if (when === 'live')
                handleImmidateErrors(_input);
        });
    }

    function handleErrors(){
        return new Promise(async(res) => {
            validating = true;
            let errors = null;
            let results = {};
            let checked = [];
            form.children().get().forEach(async(child, i, array) => {
                let field = child.find(`#${child.attr('field-id')}`);
                let sd = getSpecificDetails(field);
                let error = await checkInput(sd.info, sd.value);
                let _err = field.parent().query('.text-danger').get(0);
                if (error) {
                    field.plain(0).setCustomValidity("Error");
                    _err.refill(error);
                    errors = true;
                } else {
                    field.plain(0).setCustomValidity("");
                    _err.truncate();
                    results[sd.info.id] = sd.value;
                }
                checked.push(i)
                if(checked.length === array.length)
                   res({errors, results});
            })
        })
    }

    function inModal(){
        if(_static) return console.error('Can not perform both static and modal launch for the same Form instance');
        if(launched) return console.warn('Form instance can not be launched more than once');

        if(fields.length === 0) return console.error('formLaunchError: At least 1 field is needed to launch the form');
        let {title:tt, extraButtons:xb} = info;
        form.firstChild().removeClass('mt-2');
        let buttons = _array(xb, [xb ? xb : '']);
        buttons.push(cancelButton);
        buttons.push(submitButton);
        modal = _.modal(container, {
                backdrop: 'static',
                header: tt,
                footer: buttons,
                keyboard: false,
            });

            modal.on('hidden', () => {
                modal.drop();
                launched = false;
            });
            modal.show();
    
            cancelButton.on('click', modal.hide);
    
        _(window).on('keyup', e => {
            let tg = _(e.target);
            if ((modal ? modal.i._element : true) && e.code && e.code.toLowerCase() === 'enter' &&
                !(tg.tagName() === 'TEXTAREA' || tg.hasAttribute('contenteditable')))
                submitButton.click();
        })
        launched = true;
    }
    async function submit(){
        if(validating) return;
        loader().show()
        submitButton.blur();
        let { errors, results } = await handleErrors();
        validating = false;
        loader().hide();
        if (errors) {
            modal && modal.shake();
            form.addClass('was-validated');
            return;
        }
        modal && modal.hide();
        if(form.hasAttribute('action'))
           form.submit();
        else 
           ok && ok(results);
        reset();
    }

    function appendTo(element){
        if(fields.length === 0) return console.error('formLaunchError: At least 1 field is needed to launch the form');
        if(modal) return console.error('Can not perform both static and modal launch for the same Form instance');
        let {title, extraButtons:xb} = info;
        let buttons = _array(xb, [xb ? xb: '']);
        buttons.push(submitButton);
        container.insertBefore(__('h4').addChild(title), container.firstChild())
        container.addChild(__('text-end mt-2').addChild(buttons));
        _static = true;
        container.appendTo(element);
    }
    function exists(id) { return form.query(`#${id}`).get().plain().length > 0};
    function reset(){
        form.removeClass('was-validated').reset();
        form.query('.text-danger').get().truncate();
    }
    function close(){
        if(modal && launched)
           modal.hide();
    }
    function onOk(cb){
     if(typeof cb === 'function')
       ok = cb;
     else
       throw new Error('\'onOk\' must be a function')
    }

    return {
        /**
        * 
        * @param {[{ 
        * type: 'button' | 'checkbox' | 'textarea' | 'list' | 'color' | 'date' | 'datetime-local' | 'email' | 'file' | 'hidden' | 'image' | 'month' | 'number' | 'password' | 'radio' | 'range' | 'reset' | 'search' | 'submit' | 'tel' | 'text' | 'time' | 'url' | 'week',
        * id: Number | String,
        * placeholder,
        * default_value,
        * description: String,
        * required: Boolean,
        * label: String,
        * options: [{id: Number | String, value: Number | String, selected: Boolean}],
        * min: Number,
        * max: Number,
        * validate: {
        * validator: Function,
        * when: 'submit' | 'live'
        * }
        * }]} _fields 
        * @param {String} after 
        */
        addFields(_fields, after){return addFields(_fields, after)},
        modal(){return modal()},
        close(){return close()},
        onOk(cb){return onOk(cb)},
        exists(id){return exists(id)},
        reset(){return reset()},
        appendTo(element){return appendTo(element)},
        submit(){return submit()},
        loader(){return loader()}
    }
}